﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading;
using System.Windows.Forms;
using Emgu.CV;
using Emgu.CV.Structure;
using Emgu.Util;
using Emgu.CV.UI;
using Emgu.CV.Features2D;


namespace EmguTestApplication
{
    public partial class EmguForm : Form
    {
        struct KeyPoint
        {
            public PointF location;
            public double minEvalue;

            public KeyPoint(PointF p, double e)
            {
                location = p; minEvalue = e;
            }
        }

        private Capture sourceVideo;

        private Bitmap displayFrame;

        private PointF[] boundingBoxVertices;

        private PointF[] initialBoundingBoxVertices;

        private PointF[] displayBoundingBoxVertices;

        private MKeyPoint[] keyPoints;

        private int boundingBoxStartX = -1;

        private int boundingBoxStartY;

        private int numberOfPyramidLevels;

        private Graphics graphics;

        private Boolean mouseDown = false;
        
        private System.Windows.Forms.Timer frameGrabTimer = new System.Windows.Forms.Timer();

        private Thread trackerThread;

        private Pen pen;

        private Tracker tracker;

        public EmguForm()
        {
            InitializeComponent();

            boundingBoxVertices=new PointF[]{new PointF(-1,-1), new PointF(-1,-1), new PointF(-1,-1), new PointF(-1,-1)};

            keyPoints = new MKeyPoint[4];

            graphics = inputImagePictureBox.CreateGraphics();

            pen = new Pen(Color.Yellow);

        }

        private void initializeTrackingRegion()
        {
            Image<Bgr, byte> sourceVideoFrame = sourceVideo.QueryFrame();

            if (sourceVideoFrame == null)
                return;

            displayFrame = sourceVideoFrame.Clone().ToBitmap();

            inputImagePictureBox.Image = displayFrame;

        }

        private void debugMessageEventHandler(String message)
        {
            debugMessageLabel.Text = message;
        }

        private void frameProcessingCompleteEventHandler()
        {
            Bitmap bm = null;

            Image<Gray, byte> templateSubImage;

            // Scale the  bounding box
            float scaleFactor = (float)Math.Pow(2.0, tracker.NumberOfPyramidLevels - 1);

            float sumX = 0; 
            float sumY = 0;
            for (int i = 0; i < 4; i++)
            {
                displayBoundingBoxVertices[i].X = boundingBoxVertices[i].X * scaleFactor;
                displayBoundingBoxVertices[i].Y = boundingBoxVertices[i].Y * scaleFactor;

                sumX += displayBoundingBoxVertices[i].X;
                sumY += displayBoundingBoxVertices[i].Y;                

            }

            keyPoints[0].Point.X = sumX / 4;
            keyPoints[0].Point.Y = sumY / 4;

            if (tracker.TrackingOK)
                pen = new Pen(Color.Yellow);
            else
                pen = new Pen(Color.Red);
           

            Graphics.FromImage(tracker.DisplayFrame).DrawPolygon(pen, displayBoundingBoxVertices);

            Graphics.FromImage(tracker.DisplayFrame).FillEllipse(new SolidBrush(Color.Green), keyPoints[0].Point.X, keyPoints[0].Point.Y, 6, 6);
  
            inputImagePictureBox.Image = tracker.DisplayFrame;

            Rectangle boundingRectangle = tracker.setBoundingRectangle(displayBoundingBoxVertices);

            if ((tracker is LucasAndKanadeAffineTracker) || (tracker is LucasAndKanadeProjectiveTracker))
            {
                templateSubImage = tracker.CurrentFrame.GetSubRect(boundingRectangle);

              //  bm = new Bitmap(tracker.CurrentFrame.ToBitmap());

                bm = new Bitmap(templateSubImage.ToBitmap());

                Graphics graphics = Graphics.FromImage(bm);

                graphics.DrawPolygon(new Pen(Color.Black), displayBoundingBoxVertices);
            }
            else if (tracker is FeaturePointAffineTracker)
            {

                bm = tracker.InitialFrameSubImage.ToBitmap();         
            }

            processedImagePictureBox.Image = bm;
           
        }

        private void openFileDialog1_FileOk(object sender, CancelEventArgs e)
        {
            string fileName = openFileDialog.FileName;
            if (fileName.Length != 0)
            {
                sourceVideo = new Capture(fileName);

                // Initialize tracking region
                initializeTrackingRegion();
            }
        }

        private void fileLoadButton_Click(object sender, EventArgs e)
        {
            openFileDialog.ShowDialog();
        }
 

        private void inputImagePictureBox_MouseDown(object sender, MouseEventArgs e)
        {
            mouseDown = true;
        }

        private void inputImagePictureBox_MouseMove(object sender, MouseEventArgs e)
        {
            int endX, endY;
            if (mouseDown)
            {
                if (boundingBoxStartX == -1)
                {
                    boundingBoxStartX = e.X; boundingBoxStartY = e.Y;
                }
                else
                {
                    endX = e.X;
                    endY = e.Y;
                    graphics.DrawImage(displayFrame, 0, 0);
                    graphics.DrawRectangle(pen, new Rectangle(boundingBoxStartX, boundingBoxStartY, endX - boundingBoxStartX, endY - boundingBoxStartY));
                    boundingBoxVertices[0].X = boundingBoxStartX;
                    boundingBoxVertices[0].Y = boundingBoxStartY;
                    boundingBoxVertices[1].X = endX;
                    boundingBoxVertices[1].Y = boundingBoxStartY;
                    boundingBoxVertices[2].X = endX;
                    boundingBoxVertices[2].Y = endY;
                    boundingBoxVertices[3].X = boundingBoxStartX;
                    boundingBoxVertices[3].Y = endY;
                }
            }
        }
    
        private void inputImagePictureBox_MouseUp(object sender, MouseEventArgs e)
        {
            boundingBoxStartX = -1;
            mouseDown = false;
        }

        private void initializeTracker()
        {
            // Create initial template 
            initialBoundingBoxVertices = new PointF[4];

            displayBoundingBoxVertices = new PointF[4];

        }

     
        private void EmguForm_FormClosed(object sender, FormClosedEventArgs e)
        {
            if (trackerThread != null)
                trackerThread.Abort();
        }

        private void openToolStripMenuItem_Click(object sender, EventArgs e)
        {
            openFileDialog.ShowDialog();
        }

        private void lucasAndKanadeAffineTrackerToolStripMenuItem_Click(object sender, EventArgs e)
        {
            // Initialize tracker object
            initializeTracker();

            int maxIterations = 5;

            numberOfPyramidLevels = 2;

            // Scale the  bounding box
            float scaleFactor = (float)Math.Pow(2.0, numberOfPyramidLevels - 1);

            for (int i = 0; i < 4; i++)
            {
                boundingBoxVertices[i].X /= scaleFactor;
                boundingBoxVertices[i].Y /= scaleFactor;
                initialBoundingBoxVertices[i].X = (float)boundingBoxVertices[i].X;
                initialBoundingBoxVertices[i].Y = (float)boundingBoxVertices[i].Y;
            }

            tracker = new LucasAndKanadeAffineTracker(sourceVideo, numberOfPyramidLevels, initialBoundingBoxVertices, boundingBoxVertices, maxIterations);

            // Initialize tracker frame processing event complete handler
            tracker.FrameProcessingCompleteEvent += new frameProcessingCompleteDelegate(frameProcessingCompleteEventHandler);

            tracker.DebugMessageEvent += new debugMessageDelegate(debugMessageEventHandler);

            // Initialize tracker thread
            trackerThread = new Thread(new ThreadStart(tracker.trackRegion));

            trackerThread.Start();
        }

        private void featurePointTrackerToolStripMenuItem_Click(object sender, EventArgs e)
        {
            // Initialize tracker object
            initializeTracker();

            numberOfPyramidLevels = 1;

            // Scale the  bounding box
            float scaleFactor = (float)Math.Pow(2.0, numberOfPyramidLevels - 1);

            for (int i = 0; i < 4; i++)
            {
                boundingBoxVertices[i].X /= scaleFactor;
                boundingBoxVertices[i].Y /= scaleFactor;
                initialBoundingBoxVertices[i].X = (float)boundingBoxVertices[i].X;
                initialBoundingBoxVertices[i].Y = (float)boundingBoxVertices[i].Y;
            }

            tracker = new FeaturePointAffineTracker(sourceVideo, numberOfPyramidLevels, initialBoundingBoxVertices, boundingBoxVertices);

            // Initialize tracker frame processing event complete handler
            tracker.FrameProcessingCompleteEvent += new frameProcessingCompleteDelegate(frameProcessingCompleteEventHandler);

            tracker.DebugMessageEvent+=new debugMessageDelegate(debugMessageEventHandler);

            // Initialize tracker thread
            trackerThread = new Thread(new ThreadStart(tracker.trackRegion));

            trackerThread.Start();
        }

        private void exitToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (trackerThread != null)
                if (trackerThread.IsAlive)
                    trackerThread.Abort();

            Application.Exit();
        }

        private void EmguForm_Load(object sender, EventArgs e)
        {

        }

        private void lucasAndKanadeProjectiveTrackerToolStripMenuItem_Click(object sender, EventArgs e)
        {
            // Initialize tracker object
            initializeTracker();

            int maxIterations = 5;

            numberOfPyramidLevels = 2;

            // Scale the  bounding box
            float scaleFactor = (float)Math.Pow(2.0, numberOfPyramidLevels - 1);

            for (int i = 0; i < 4; i++)
            {
                boundingBoxVertices[i].X /= scaleFactor;
                boundingBoxVertices[i].Y /= scaleFactor;
                initialBoundingBoxVertices[i].X = (float)boundingBoxVertices[i].X;
                initialBoundingBoxVertices[i].Y = (float)boundingBoxVertices[i].Y;
            }

            tracker = new LucasAndKanadeProjectiveTracker(sourceVideo, numberOfPyramidLevels, initialBoundingBoxVertices, boundingBoxVertices, maxIterations);

            // Initialize tracker frame processing event complete handler
            tracker.FrameProcessingCompleteEvent += new frameProcessingCompleteDelegate(frameProcessingCompleteEventHandler);

            tracker.DebugMessageEvent += new debugMessageDelegate(debugMessageEventHandler);

            // Initialize tracker thread
            trackerThread = new Thread(new ThreadStart(tracker.trackRegion));

            trackerThread.Start();
        }
    }
}
