﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Drawing;
using System.Threading;
using Emgu.CV;
using Emgu.CV.Util;
using Emgu.CV.Structure;
using Emgu.CV.CvEnum;
using Emgu.Util;
using Emgu.CV.Features2D;
using System.Runtime.InteropServices;

namespace EmguTestApplication
{
    public delegate void frameProcessingCompleteDelegate();

    public delegate void debugMessageDelegate(String message);

    
    class MatrixOps
    {
        public static void copyMatrix(double[,] from, double[,] to)
        {
            for (int i = 0; i < from.GetLength(0); i++)
                for (int j = 0; j < from.GetLength(1); j++)
                    to[i, j] = from[i, j];
        }

        public static double[,] matrixTranspose(double[,] mat)
        {
            int nrows = mat.GetLength(0); int ncols = mat.GetLength(1);
            double[,] mat1 = new double[ncols, nrows];
            for (int i = 0; i < nrows; i++)
                for (int j = 0; j < ncols; j++)
                    mat1[j, i] = mat[i, j];

            return mat1;
        }

        public static void matrixSum(double[,] mat, double[,] mat1)
        {
            int nrows = mat.GetLength(0); int ncols = mat.GetLength(1);
            for (int i = 0; i < nrows; i++)
                for (int j = 0; j < ncols; j++)
                    mat[i, j] += mat1[i, j];
        }


        public static double[,] matrixMult(double[,] mat1, double[,] mat2)
        {
            int nrows = mat1.GetLength(0); int ncols = mat2.GetLength(1);
            int np1 = mat1.GetLength(1); int np2 = mat2.GetLength(0);
            if (np1 != np2)
                return null;
            double[,] mat = new double[nrows, ncols];
            for (int i = 0; i < nrows; i++)
                for (int j = 0; j < ncols; j++)
                    mat[i, j] = 0.0;

            for (int i = 0; i < nrows; i++)
            {
                for (int j = 0; j < ncols; j++)
                {
                    double sum = 0;
                    for (int k = 0; k < np1; k++)
                        sum = sum + mat1[i, k] * mat2[k, j];

                    mat[i, j] = sum;
                }
            }

            return mat;
        }

        public static double[] matrixVecMult(double[,] mat, double[] vec)
        {
            int nrows = mat.GetLength(0);
            int n = vec.Length;
            double[] productVec = new double[nrows];

            for (int i = 0; i < nrows; i++)
                productVec[i] = 0.0;

            for (int i = 0; i < nrows; i++)
                for (int j = 0; j < n; j++)
                    productVec[i] += (mat[i, j] * vec[j]);

            return productVec;
        }

        public static double minimumEigenValue(double[,] mat)
        {
            // Returns the minimum eigenvalue of a symmetrix matrix
            return (0.5 * (mat[0,0] + mat[1,1] - Math.Sqrt(4 * mat[0,1] * mat[1,0] + (mat[0,0] - mat[1,1]) * (mat[0,0] - mat[1,1]))));
        }

        public static void linearLeastSquares(double[,] mat, double[] vec, double[] paramVec)
        {
            // Solves the least squares problem AX=b
            int info;
            alglib.lsfitreport rep;

            //
            // Linear fitting without weights
            //
            alglib.lsfitlinear(vec, mat, out info, out paramVec, out rep);
        }
    }

    class Tracker
    {
        private event frameProcessingCompleteDelegate frameProcessingCompleteEvent;

        private event debugMessageDelegate debugMessageEvent;

        protected int numberOfPyramidLevels = 1;

        protected Capture sourceVideo;

        protected PointF[] initialBoundingBoxVertices, boundingBoxVertices;

        protected Bitmap displayFrame;

        protected Image<Bgr, byte> sourceVideoFrame;

        protected Image<Gray, byte> currentFrame, previousFrame;

        protected Image<Gray, byte> initialFrame;

        protected Image<Gray, byte> initialFrameSubImage;

        protected Boolean trackingOK;

        protected double boundingBoxDisplacementThreshold;

        protected int keyPointFeatureWindowSize;            // Window size for key points

        protected int searchLength;                         // Search length in x and y to maximize interest measure                

        protected ArrayList keyPoints;

        protected FastDetector fastDetector;

        protected struct KeyPoint
        {
            public PointF location;
            public double minEvalue;

            public KeyPoint(PointF p, double e)
            {
                location = p; minEvalue = e;
            }
        }

        protected Boolean insidePolygon(int x, int y, PointF[] bBoxVertices)
        {
            // Compute the cross product test
            int sgnPrev = 0;
            Boolean interior = true;
            for (int v = 0; v <= 3; v++)
            {
                int xdiff = (x - (int)bBoxVertices[v].X);
                int ydiff = (y - (int)bBoxVertices[v].Y);
                int x1diff = ((int)bBoxVertices[v].X - (int)bBoxVertices[(v + 1) % 4].X);
                int y1diff = ((int)bBoxVertices[v].Y - (int)bBoxVertices[(v + 1) % 4].Y);

                int sgn = Math.Sign((ydiff * x1diff) - (xdiff * y1diff));

                if (sgn == 0)
                {
                    interior = false;
                    break;
                }

                if (sgnPrev != 0)
                {
                    if (sgn * sgnPrev < 0)
                    {
                        interior = false;
                        break;
                    }
                }

                sgnPrev = sgn;
            }

            return interior;
        }

        protected void computeStatistics(PointF[] boundingBoxVertices, Image<Gray, byte> image, out double mean, out double sd)
        {
            double sum = 0.0;
            double sumsq = 0.0;
            double grey = 0.0;
            double npixels = 0;

            // Determine bounding box rectangle
            Rectangle boundingRectangle = setBoundingRectangle(boundingBoxVertices);

            byte[, ,] data = image.Data;
            for (int x = boundingRectangle.X; x < boundingRectangle.X + boundingRectangle.Width; x++)
                for (int y = boundingRectangle.Y; y < boundingRectangle.Y + boundingRectangle.Height; y++)
                {
                    if (insidePolygon(x, y, boundingBoxVertices))
                    {
                        if ((x >= 0) && (x < image.Width))
                            if ((y >= 0) && (y < image.Height))
                            {
                                grey = data[y, x, 0];
                                sum += grey;
                                sumsq += (grey * grey);
                                npixels++;
                            }
                    }
                }
            mean = sum / npixels;
            sd = Math.Sqrt(sumsq / npixels - (mean * mean));
        }

        protected void updateWarp(double[,] affineWarpMatrixOld, double[] displacementVectorOld, double[,] affineWarpMatrix, double[] displacmentVector)
        {
            // Affine matrix and displacementvector overwritten
            double[] tempVector = MatrixOps.matrixVecMult(affineWarpMatrix, displacementVectorOld);

            double[,] tempMatrix = MatrixOps.matrixMult(affineWarpMatrix, affineWarpMatrixOld);

            MatrixOps.copyMatrix(tempMatrix, affineWarpMatrixOld);

            displacementVectorOld[0] = tempVector[0] + displacmentVector[0];
            displacementVectorOld[1] = tempVector[1] + displacmentVector[1];
        }

        protected void updateWarp(double[,] projectiveParameterMatrixOld, double[,] projectiveParameterMatrix)
        {
            double scale = projectiveParameterMatrixOld[0, 2] * projectiveParameterMatrix[2, 0] + projectiveParameterMatrixOld[1, 2] * projectiveParameterMatrix[2, 1] + 1;

            double[,] updatedProjectiveParameterMatrix = MatrixOps.matrixMult(projectiveParameterMatrix, projectiveParameterMatrixOld);

            for (int i = 0; i < 3; i++)
                for (int j = 0; j < 3; j++)
                    projectiveParameterMatrixOld[i, j] = updatedProjectiveParameterMatrix[i, j] / scale;
        }
            

        protected double boundingBoxDisplacement(PointF[] bbox1, PointF[] bbox2)
        {
            // Computes the maximum displacment between the vertices of 2 bounding boxes

            double maxDisplacement = 0.0;
            for (int i = 0; i < 4; i++)
            {
                double d = Math.Sqrt((bbox1[i].X - bbox2[i].X) * (bbox1[i].X - bbox2[i].X) + (bbox1[i].Y - bbox2[i].Y) * (bbox1[i].Y - bbox2[i].Y));

                if (d > maxDisplacement)
                    maxDisplacement = d;
            }

            return maxDisplacement;
        }

        protected double[] imageGradient(Image<Gray, byte> image, Point p)
        {
            double[] gradient = new double[2];
            gradient[0] = gradient[1] = 0.0;
            if ((p.X < 1) || (p.X >= (image.Width - 1)))
                return gradient;

            if ((p.Y < 1) || (p.Y >= (image.Height - 1)))
                return gradient;

            gradient[0] = (image.Data[p.Y, p.X + 1, 0] - image.Data[p.Y, p.X - 1, 0]) / 2.0;

            gradient[1] = (image.Data[p.Y + 1, p.X, 0] - image.Data[p.Y - 1, p.X, 0]) / 2.0;

            return gradient;
        }

        protected double[] imageGradient(Image<Gray, byte> image, PointF p)
        {
            double[] gradient = new double[2];

            gradient[0] = gradient[1] = 0.0;
            if ((p.X < 1) || (p.X >= (image.Width - 1)))
                return gradient;

            if ((p.Y < 1) || (p.Y >= (image.Height - 1)))
                return gradient;

            double gx1 = biLinearInterpolate(image, p.X + 1, p.Y);
            double gx2 = biLinearInterpolate(image, p.X - 1, p.Y);
            double gy1 = biLinearInterpolate(image, p.X, p.Y + 1);
            double gy2 = biLinearInterpolate(image, p.X, p.Y - 1);

            gradient[0] = (gx1 - gx2) / 2;
            gradient[1] = (gy1 - gy2) / 2;


            return gradient;
        }

        public Rectangle setBoundingRectangle(PointF[] bBoxVertices)
        {
            double xmin = 1.0e10;
            xmin = Math.Min(xmin, bBoxVertices[0].X);
            xmin = Math.Min(xmin, bBoxVertices[1].X);
            xmin = Math.Min(xmin, bBoxVertices[2].X);
            xmin = Math.Min(xmin, bBoxVertices[3].X);
            double xmax = -1.0e10;
            xmax = Math.Max(xmax, bBoxVertices[0].X);
            xmax = Math.Max(xmax, bBoxVertices[1].X);
            xmax = Math.Max(xmax, bBoxVertices[2].X);
            xmax = Math.Max(xmax, bBoxVertices[3].X);

            double ymin = 1.0e10;
            ymin = Math.Min(ymin, bBoxVertices[0].Y);
            ymin = Math.Min(ymin, bBoxVertices[1].Y);
            ymin = Math.Min(ymin, bBoxVertices[2].Y);
            ymin = Math.Min(ymin, bBoxVertices[3].Y);
            double ymax = -1.0e10;
            ymax = Math.Max(ymax, bBoxVertices[0].Y);
            ymax = Math.Max(ymax, bBoxVertices[1].Y);
            ymax = Math.Max(ymax, bBoxVertices[2].Y);
            ymax = Math.Max(ymax, bBoxVertices[3].Y);

            Rectangle boundingRectangle = new Rectangle((int)xmin, (int)ymin, (int)(xmax - xmin + 0.5), (int)(ymax - ymin + 0.5));

            return boundingRectangle;
        }

        protected double[] warpPoint(PointF p, double[,] affineWarpMatrix, double[] displacmentVector)
        {
            double[] warpedPoint = new double[2];
            warpedPoint[0] = affineWarpMatrix[0, 0] * p.X + affineWarpMatrix[0, 1] * p.Y + displacmentVector[0];
            warpedPoint[1] = affineWarpMatrix[1, 0] * p.X + affineWarpMatrix[1, 1] * p.Y + displacmentVector[1];

            return warpedPoint;
        }

        protected double[] warpPoint(PointF p, double[,] projectiveParameterMatrix)
        {
            double[] warpedPoint = new double[2];

            double den = projectiveParameterMatrix[2, 0] * p.X + projectiveParameterMatrix[2, 1] * p.Y + 1;
            warpedPoint[0] = (projectiveParameterMatrix[0, 0] * p.X + projectiveParameterMatrix[0, 1] * p.Y + projectiveParameterMatrix[0, 2]) / den;
            warpedPoint[1] = (projectiveParameterMatrix[1, 0] * p.X + projectiveParameterMatrix[1, 1] * p.Y + projectiveParameterMatrix[1, 2]) / den;

            return warpedPoint;
        }

        protected double biLinearInterpolate(Image<Gray, byte> image, double x, double y)
        {
            if (((int)x >= (image.Width - 1)) || ((int)y >= (image.Height - 1)))
                return image.Data[image.Height - 1, image.Width - 1, 0];

            if (((int)x < 0) || ((int)y < 0))
                return image.Data[0, 0, 0];

            int x1 = (int)x; int x2 = (x1 + 1);
            int y1 = (int)y; int y2 = (y1 + 1);

            byte[, ,] data = image.Data;

            double gray11 = data[y1, x1, 0];
            double gray12 = data[y2, x1, 0];
            double gray21 = data[y1, x2, 0];
            double gray22 = data[y2, x2, 0];

            return (gray11 * (x2 - x) * (y2 - y) + gray12 * (x2 - x) * (y - y1) + gray21 * (x - x1) * (y2 - y) + gray22 * (x - x1) * (y - y1));
        }

        protected void warpBoundingBox(double[,] affineWarpMatrix, double[] displacementVector, PointF[] bbVertices)
        {
            PointF[] vertices = new PointF[4];
            for (int i = 0; i < 4; i++)
            {
                vertices[i].X = (float)(warpPoint(bbVertices[i], affineWarpMatrix, displacementVector)[0]);
                vertices[i].Y = (float)(warpPoint(bbVertices[i], affineWarpMatrix, displacementVector)[1]);
            }

            for (int i = 0; i < 4; i++)
                bbVertices[i] = vertices[i];
        }

        protected void warpBoundingBox(double[,] projectiveParameterMatrix, PointF[] bbVertices)
        {
            PointF[] vertices = new PointF[4];
            for (int i = 0; i < 4; i++)
            {
                vertices[i].X = (float)(warpPoint(bbVertices[i], projectiveParameterMatrix)[0]);
                vertices[i].Y = (float)(warpPoint(bbVertices[i], projectiveParameterMatrix)[1]);
            }

            for (int i = 0; i < 4; i++)
                bbVertices[i] = vertices[i];
        }

        protected double[,] computeOrientationMatrix(Image<Gray, Byte> frame, int x, int y, int windowSize, double gaussianSigma)
        {
            double[] grad;
            double gauss, r2;
            double ixx = 0.0; double ixy = 0.0; double iyy = 0.0;
            for (int x1 = -windowSize / 2; x1 <= windowSize / 2; x1++)
                for (int y1 = -windowSize / 2; y1 <= windowSize / 2; y1++)
                {
                    int x2 = x + x1; int y2 = y + y1;
                    if ((x2 > 0) && (x2 < (frame.Width - 1)))
                        if ((y2 > 0) && (y2 < (frame.Height - 1)))
                        {
                            grad = imageGradient(frame, new Point(x2, y2));
                            r2 = (x1 * x1) + (y1 * y1);
                            gauss = Math.Exp(-r2 / (2.0 * gaussianSigma));
                            gauss = 1.0;
                            ixx += (gauss * grad[0] * grad[0]);
                            iyy += (gauss * grad[1] * grad[1]);
                            ixy += (gauss * grad[0] * grad[1]);
                        }
                }

            double[,] orientationMatrix = { { ixx, ixy }, { ixy, iyy } };

            return orientationMatrix;
        }

        protected ArrayList getKeyPoints(Image<Gray, Byte> frame, PointF[] bBoxVertices, int searchLength, int windowSize)
        {
            keyPoints = new ArrayList();

            Rectangle roi = setBoundingRectangle(bBoxVertices);

            double maxMinValue = 0;
            for (int x = roi.X + searchLength; x < roi.Width + roi.X - searchLength; x += searchLength)
                for (int y = roi.Y + searchLength; y < roi.Height + roi.Y - searchLength; y += searchLength)
                {
                    // Compute the maximum of the minimum eigenvalues within each feature window
                    double maxEvalue = 0;
                    int xi = 0;
                    int yi = 0;
                    for (int x1 = 0; x1 < searchLength; x1++)
                        for (int y1 = 0; y1 < searchLength; y1++)
                        {
                            int x2 = x + x1;
                            int y2 = y + y1;
                            if ((x2 >= 0) && (x2 < (frame.Width)))
                                if ((y2 >= 0) && (y2 < (frame.Height)))
                                {
                                    if (insidePolygon(x2, y2, bBoxVertices))
                                    {
                                        double emin = MatrixOps.minimumEigenValue(computeOrientationMatrix(frame, x2, y2, windowSize, 1.0));
                                        if (emin > maxEvalue)
                                        {
                                            maxEvalue = emin;
                                            xi = x2;
                                            yi = y2;
                                        }
                                    }
                                }
                        }

                    if (maxEvalue > 0)
                    {
                        KeyPoint kp = new KeyPoint(new PointF(xi, yi), maxEvalue);

                        PointF pp = new PointF(xi, yi);

                        keyPoints.Add(kp);

                        if (maxEvalue > maxMinValue)
                            maxMinValue = maxEvalue;
                    }
                }

            // Filter keyPoints
            for (int i = 0; i < keyPoints.Count; i++)
            {
                KeyPoint k = (KeyPoint)keyPoints[i];
                if (k.minEvalue < (0.75 * maxMinValue))
                {
                    keyPoints.RemoveAt(i);
                }

            }

            return keyPoints;
        }


        protected void triggerNextFrameEvent()
        {
            frameProcessingCompleteEvent();
        }

        protected void triggerDebugMessageEvent(String message)
        {
            debugMessageEvent(message);
        }

        public Bitmap DisplayFrame
        {
            get { return displayFrame; }
        }

        public Image<Gray, Byte> InitialFrameSubImage
        {
            get { return initialFrameSubImage; }
        }

      
        public ArrayList KeyPoints
        {
            get
            {
                return keyPoints;
            }
        }

        public Image<Gray,Byte>  CurrentFrame
        {
            get { return currentFrame; }
        }

  
        public event frameProcessingCompleteDelegate FrameProcessingCompleteEvent
        {
            add
            {
                frameProcessingCompleteEvent += value;
            }
            remove
            {
                frameProcessingCompleteEvent -= value;
            }
        }

        public event debugMessageDelegate DebugMessageEvent
        {
            add
            {
                debugMessageEvent += value;
            }
            remove
            {
                debugMessageEvent -= value;
            }
        }

        public Boolean TrackingOK
        {
            get { return trackingOK; }
        }

        public int NumberOfPyramidLevels
        {
            get { return numberOfPyramidLevels; }
        }

        public Tracker(Capture source, int nLevels, PointF[] initialVertices, PointF[] vertices)
        {
            sourceVideo = source;

            previousFrame = null;

            initialBoundingBoxVertices = initialVertices;

            boundingBoxVertices = vertices;

            boundingBoxDisplacementThreshold = 3;


            fastDetector = new FastDetector(10, true);
        }

        public virtual void trackRegion()
        {         
        }
    }

    class FeaturePointAffineTracker : Tracker
    {
        private int correlationFeatureWindowSize;       // Window size for correlation matching

        private int searchRadius;                       // Search radius for correlation matching   
           
        private PointF computeMaxCorrelationPoint(Image<Gray, Byte> frame1, Image<Gray, Byte> frame2, PointF[] boundingBoxVertices, int x, int y, int windowSize, int searchRadius)
        {
            int xi = 0;
            int yi = 0;
            double maxCorr = 0.0;
            
            for (int x1 = x - searchRadius; x1 <= x + searchRadius; x1++)
                for (int y1 = y - searchRadius; y1 <= y + searchRadius; y1++)
                {
                    double sumCorr = 0.0;
                    double sumSqInt1 = 0.0;
                    double sumSqInt2 = 0.0;
                    double sumInt1 = 0.0;
                    double sumInt2 = 0.0;
                    double count = 0;
                    double corr = 0.0;
                    if (insidePolygon(x1, y1, boundingBoxVertices))
                    {
                        for (int x2 = -windowSize/2; x2 <= windowSize/2; x2++)
                            for (int y2 = -windowSize / 2; y2 <= windowSize / 2; y2++)
                            {
                                int x3 = x + x2;
                                int y3 = y + y2;
                                int x4 = x1 + x2;
                                int y4 = y1 + y2;
                                if ((x3 >= 0) && (x3 < (frame1.Width)))
                                    if ((y3 >= 0) && (y3 < (frame1.Height)))
                                        if ((x4 >= 0) && (x4 < (frame2.Width)))
                                            if ((y4 >= 0) && (y4 < (frame2.Height)))
                                            {
                                                sumInt1 += (frame1[y3, x3].Intensity);
                                                sumInt2 += (frame2[y4, x4].Intensity);
                                                count++;
                                            }
                            }
                        double mean1 = sumInt1 / count;
                        double mean2 = sumInt2 / count;
                        for (int x2 = -windowSize/2; x2 <= windowSize/2; x2++)
                            for (int y2 = -windowSize/2; y2 <= windowSize/2; y2++)
                            {
                                int x3 = x + x2;
                                int y3 = y + y2;
                                int x4 = x1 + x2;
                                int y4 = y1 + y2;
                                if ((x3 >= 0) && (x3 < (frame1.Width)))
                                    if ((y3 >= 0) && (y3 < (frame1.Height)))
                                        if ((x4 >= 0) && (x4 < (frame2.Width)))
                                            if ((y4 >= 0) && (y4 < (frame2.Height)))
                                            {
                                                sumCorr += (frame1[y3, x3].Intensity - mean1) * (frame2[y4, x4].Intensity - mean2);
                                                sumSqInt1 += (frame1[y3, x3].Intensity - mean1) * (frame1[y3, x3].Intensity - mean1);
                                                sumSqInt2 += (frame2[y4, x4].Intensity - mean2) * (frame2[y4, x4].Intensity - mean2);
                                            }
                            }

                        if ((sumInt1 > 0) && (sumInt2 > 0))
                           corr = sumCorr / Math.Sqrt(sumSqInt1 * sumSqInt2);

                        if (corr > maxCorr)
                        {
                            maxCorr = corr;
                            xi = x1;
                            yi = y1;
                        }
                    }
                }

            return new PointF(xi,yi);
        }

        private void align(PointF[] templateBoundingBoxVertices, Image<Gray, Byte> templateFrame, Image<Gray, Byte> currFrame,
                          double[,] startAffineWarpMatrix, double[] startDisplacementVector, double[,] affineWarpMatrix, double[] displacementVector, int lab)
        {
            ArrayList keyPoints;
            double[,] dataMatrix;
            double[] dataVector;
            double[] affineParameterVector = new double[6];
            int info;
            Rectangle templateBoundingRectangle;
            Image<Gray, Byte> affineWarpedFrame;
            alglib.lsfitreport rep;

            // Determine bounding box rectangle
            templateBoundingRectangle = setBoundingRectangle(templateBoundingBoxVertices);


            // Create image to hold warped frame
            affineWarpedFrame = new Image<Gray, Byte>(currFrame.Width, currFrame.Height);

            // Copy the initial warp parameters
            MatrixOps.copyMatrix(startAffineWarpMatrix, affineWarpMatrix);
            displacementVector[0] = startDisplacementVector[0];
            displacementVector[1] = startDisplacementVector[1];

 
            for (int x = templateBoundingRectangle.X - searchRadius; x <= templateBoundingRectangle.X + templateBoundingRectangle.Width + searchRadius; x++)
                for (int y = templateBoundingRectangle.Y - searchRadius; y <= templateBoundingRectangle.Y + templateBoundingRectangle.Height + searchRadius; y++)
                {
                    if ((x >= 0) && (x < affineWarpedFrame.Width) && (y >= 0) && (y < affineWarpedFrame.Height))
                    {
                        double warpX = warpPoint(new PointF(x, y), affineWarpMatrix, displacementVector)[0];
                        double warpY = warpPoint(new PointF(x, y), affineWarpMatrix, displacementVector)[1];

                        affineWarpedFrame[y, x] = new Gray((int)biLinearInterpolate(currFrame, warpX, warpY));
                    }
                }

            
            keyPoints = getKeyPoints(affineWarpedFrame, templateBoundingBoxVertices, searchLength, keyPointFeatureWindowSize);

            dataMatrix = new double[2 * keyPoints.Count, 6];

            dataVector = new double[2 * keyPoints.Count];

            int rowIndex = 0;

            foreach (KeyPoint p in keyPoints)
            {
                PointF p1 = computeMaxCorrelationPoint(templateFrame, affineWarpedFrame, templateBoundingBoxVertices, (int)p.location.X, (int)p.location.Y, correlationFeatureWindowSize, searchRadius);

                dataVector[rowIndex] = p1.X;
                dataVector[rowIndex + 1] = p1.Y;

                dataMatrix[rowIndex, 0] = p.location.X;
                dataMatrix[rowIndex, 1] = p.location.Y;
                dataMatrix[rowIndex, 2] = 0;
                dataMatrix[rowIndex, 3] = 0;
                dataMatrix[rowIndex, 4] = 1;
                dataMatrix[rowIndex, 5] = 0;
                dataMatrix[rowIndex + 1, 0] = 0;
                dataMatrix[rowIndex + 1, 1] = 0;
                dataMatrix[rowIndex + 1, 2] = p.location.X;
                dataMatrix[rowIndex + 1, 3] = p.location.Y;
                dataMatrix[rowIndex + 1, 4] = 0;
                dataMatrix[rowIndex + 1, 5] = 1;

                rowIndex += 2;
            }

            alglib.lsfitlinear(dataVector, dataMatrix, out info, out affineParameterVector, out rep);

            // Return the total affine matrix and displacement vector (including the  starting value)

            affineWarpMatrix[0, 0] = affineParameterVector[0];
            affineWarpMatrix[0, 1] = affineParameterVector[1];
            affineWarpMatrix[1, 0] = affineParameterVector[2];
            affineWarpMatrix[1, 1] = affineParameterVector[3];

            displacementVector[0] = affineParameterVector[4];
            displacementVector[1] = affineParameterVector[5];

            updateWarp(affineWarpMatrix, displacementVector, startAffineWarpMatrix, startDisplacementVector);
        }

        public FeaturePointAffineTracker(Capture source, int nLevels, PointF[] initialVertices, PointF[] vertices)
            : base(source, nLevels, initialVertices, vertices)
        {

            keyPointFeatureWindowSize = 5;

            correlationFeatureWindowSize = 5;

            searchRadius = 5;

            searchLength = 5;

            trackingOK = true;
        }

        public override void trackRegion()
        {
            double[,] affineWarpMatrix = new double[2, 2];          // Frame to frame warp matrix
            double[] displacementVector = new double[2];            // Frame to frame displacement
            double[,] startAffineWarpMatrix = new double[2, 2];     // Frame to frame warp matrix
            double[] startDisplacementVector = new double[2];       // Frame to frame displacement
            double[,] totalAffineWarpMatrix = new double[2, 2];     // Frame 0 to current frame warp matrix
            double[] totalDisplacementVector = new double[2];       // Frame 0 to current frame displacement
            PointF[] warpedBoundingBoxVertices = new PointF[4];
            double maxDisp;

            sourceVideoFrame = sourceVideo.QueryFrame();

            // Initialize the total affine parameters
            totalAffineWarpMatrix[0, 0] = totalAffineWarpMatrix[1, 1] = 1.0;
            totalAffineWarpMatrix[1, 0] = totalAffineWarpMatrix[0, 1] = 0.0;
            totalDisplacementVector[0] = totalDisplacementVector[1] = 0.0;

            while (sourceVideoFrame != null)
            {

                // Convert to grayscale
                currentFrame = sourceVideoFrame.Convert<Gray, Byte>();

                if (initialFrame == null)
                {
                    initialFrame = currentFrame.Clone();

                    Rectangle initialBoundingRect = setBoundingRectangle(initialBoundingBoxVertices);

                    initialFrameSubImage = initialFrame.GetSubRect(initialBoundingRect);
                }

                if (previousFrame != null)
                {
                    displayFrame = sourceVideoFrame.Clone().ToBitmap();

                    // Initialize the affine parameters
                    startAffineWarpMatrix[0, 0] = startAffineWarpMatrix[1, 1] = 1.0;
                    startAffineWarpMatrix[1, 0] = startAffineWarpMatrix[0, 1] = 0.0;
                    startDisplacementVector[0] = startDisplacementVector[1] = 0.0;

                    // Perform the LK alignment
                    align(boundingBoxVertices, previousFrame, currentFrame, startAffineWarpMatrix, startDisplacementVector, affineWarpMatrix, displacementVector,0);

                   // Update the bounding box vertices according to the affine warp matrix and displacement
                    for (int i = 0; i < 4; i++)
                        warpedBoundingBoxVertices[i] = initialBoundingBoxVertices[i];

                    // Update total affine warp parameters
                    updateWarp(totalAffineWarpMatrix, totalDisplacementVector, affineWarpMatrix, displacementVector);

                    // Warp bounding box
                    warpBoundingBox(totalAffineWarpMatrix, totalDisplacementVector, warpedBoundingBoxVertices);

                    // Starting from initial bounding box and current total affine parameter estimates, align initial frame
                    MatrixOps.copyMatrix(totalAffineWarpMatrix, startAffineWarpMatrix);
                    startDisplacementVector[0] = totalDisplacementVector[0];
                    startDisplacementVector[1] = totalDisplacementVector[1];

                    align(initialBoundingBoxVertices, initialFrame, currentFrame, startAffineWarpMatrix, startDisplacementVector, affineWarpMatrix, displacementVector,1);

                    for (int i = 0; i < 4; i++)
                        boundingBoxVertices[i] = initialBoundingBoxVertices[i];

                    warpBoundingBox(affineWarpMatrix, displacementVector, boundingBoxVertices);

                    maxDisp = boundingBoxDisplacement(boundingBoxVertices, warpedBoundingBoxVertices);

                    triggerDebugMessageEvent("maxDisp= " + maxDisp);


                    // Test for consistent bounding box  positions
                    if (maxDisp < boundingBoxDisplacementThreshold)
                    {
                        // Update total affine warp parameters
                        MatrixOps.copyMatrix(affineWarpMatrix, totalAffineWarpMatrix);
                        totalDisplacementVector[0] = displacementVector[0];
                        totalDisplacementVector[1] = displacementVector[1];

                        trackingOK = true;
                    }
                    else
                    {
                        trackingOK = false;

                        // Restore bounding box
                        for (int i = 0; i < 4; i++)
                            boundingBoxVertices[i] = warpedBoundingBoxVertices[i];
                    }  

                    // Signal for next frame
                    triggerNextFrameEvent();
                }

                // Copy current frame
                previousFrame = currentFrame.Copy();

                // Get next frame
                sourceVideoFrame = sourceVideo.QueryFrame();
                
            }
                            
        }
    }

    class LucasAndKanadeAffineTracker : Tracker
    {
        private ImagePyramid currentFramePyramid, previousFramePyramid, initialFramePyramid;

        private Image<Gray, byte> warpedFrame;

        private int maxIterationCount;

        private void align(PointF[] templateBoundingBoxVertices, Image<Gray, Byte> templateFrame, Image<Gray, Byte> currFrame,
                           double[,] startAffineWarpMatrix, double[] startDisplacementVector, double[,] affineWarpMatrix, double[] displacementVector)
        {
            alglib.matinvreport rep;
            int info;
            Rectangle templateBoundingRectangle;
            double[,] sumSpatialGradientMatrix = new double[6, 6];
            double[,] spatialGradientMatrix = null;
            double[,] tempMatrix = new double[2, 2];
            double[] frameMismatchVector = new double[6];
            double[] LKParameterVector = new double[6];
            double warpedFrameMean, warpedFrameSD;
            double templateFrameMean, templateFrameSD;
            double contrastScale, brightnessScale;
            double[,] structureMatrix = new double[2, 2];
            double[,] jacobianMatrix = new double[2, 6];
            double[,] tempMatrix1, tempMatrix2;
            ArrayList keyPoints;
  

            // Create image to hold warped frame
            Image<Gray,Byte> affineWarpedFrame = new Image<Gray, Byte>(currFrame.Width, currFrame.Height);
         
            // Determine bounding box rectangle
            templateBoundingRectangle = setBoundingRectangle(templateBoundingBoxVertices);

            // Access pixels in current bounding polygon
            // Iterate over each point in the outer bounding box
            // Compute the spatial gradient matrix

            for (int i = 0; i < 6; i++)
                for (int j = 0; j < 6; j++)
                    sumSpatialGradientMatrix[i, j] = 0.0;


            for (int x = templateBoundingRectangle.X; x < templateBoundingRectangle.X + templateBoundingRectangle.Width; x++)
                for (int y = templateBoundingRectangle.Y; y < templateBoundingRectangle.Y + templateBoundingRectangle.Height; y++)
                {
                    if (insidePolygon(x, y, templateBoundingBoxVertices))
                    {
                        if ((x >= 0) && (x < affineWarpedFrame.Width) && (y >= 0) && (y < affineWarpedFrame.Height))
                        {                       
                            double[] grad = imageGradient(templateFrame, new Point(x,y));

                            structureMatrix[0, 0] = grad[0] * grad[0];
                            structureMatrix[0, 1] = grad[0] * grad[1];
                            structureMatrix[1, 0] = grad[1] * grad[0];
                            structureMatrix[1, 1] = grad[1] * grad[1];

                            jacobianMatrix[0, 0] = 1; jacobianMatrix[1, 0] = 0;
                            jacobianMatrix[0, 1] = 0; jacobianMatrix[1, 1] = 1;
                            jacobianMatrix[0, 2] = x; jacobianMatrix[1, 2] = 0;
                            jacobianMatrix[0, 3] = y; jacobianMatrix[1, 3] = 0;
                            jacobianMatrix[0, 4] = 0; jacobianMatrix[1, 4] = x;
                            jacobianMatrix[0, 5] = 0; jacobianMatrix[1, 5] = y;

                            tempMatrix1 = MatrixOps.matrixMult(structureMatrix, jacobianMatrix);

                            tempMatrix2 = MatrixOps.matrixTranspose(jacobianMatrix);

                            spatialGradientMatrix = MatrixOps.matrixMult(tempMatrix2, tempMatrix1);

                            MatrixOps.matrixSum(sumSpatialGradientMatrix, spatialGradientMatrix);

                        }
                    }
                } 

            // Invert the spatial gradient matrix
            alglib.rmatrixinverse(ref sumSpatialGradientMatrix, out info, out rep);

            // Copy the initial warp parameters
            MatrixOps.copyMatrix(startAffineWarpMatrix, affineWarpMatrix);
            displacementVector[0] = startDisplacementVector[0];
            displacementVector[1] = startDisplacementVector[1];

            // Start the LK iterative improvement
            int iterationCount = 0;
            do
            {

                // Warp frame according to current parameter estimates
                for (int x = templateBoundingRectangle.X; x < templateBoundingRectangle.X + templateBoundingRectangle.Width ; x++)
                    for (int y = templateBoundingRectangle.Y; y < templateBoundingRectangle.Y + templateBoundingRectangle.Height; y++)
                    {
                        if (insidePolygon(x, y, templateBoundingBoxVertices))
                        {
                            if ((x >= 0) && (x < affineWarpedFrame.Width) && (y >= 0) && (y < affineWarpedFrame.Height))
                            {

                                double warpX = warpPoint(new PointF(x, y), affineWarpMatrix, displacementVector)[0];
                                double warpY = warpPoint(new PointF(x, y), affineWarpMatrix, displacementVector)[1];

                                affineWarpedFrame[y, x] = new Gray((int)biLinearInterpolate(currFrame, warpX, warpY));

                            }
                        }
                    }

         

                // Normalise the mean luminance and contrast of the warped frame to match  the  template frame
                computeStatistics(templateBoundingBoxVertices, templateFrame, out templateFrameMean, out templateFrameSD);

                computeStatistics(templateBoundingBoxVertices, affineWarpedFrame, out warpedFrameMean, out warpedFrameSD);

                contrastScale = templateFrameSD/warpedFrameSD;

                brightnessScale = templateFrameMean - contrastScale * warpedFrameMean;

                contrastScale = 1; brightnessScale = 0;

               // Compute the mismatch vector
                for (int i = 0; i < 6; i++)
                    frameMismatchVector[i] = 0.0;
                
                for (int x = templateBoundingRectangle.X; x < templateBoundingRectangle.X + templateBoundingRectangle.Width; x++)
                    for (int y = templateBoundingRectangle.Y; y < templateBoundingRectangle.Y + templateBoundingRectangle.Height; y++)
                    {
                        if (insidePolygon(x, y, templateBoundingBoxVertices))
                        {
                            if ((x >= 0) && (x < affineWarpedFrame.Width) && (y >= 0) && (y < affineWarpedFrame.Height))
                            {
                                double warpX = warpPoint(new PointF(x, y), affineWarpMatrix, displacementVector)[0];
                                double warpY = warpPoint(new PointF(x, y), affineWarpMatrix, displacementVector)[1];
                                double[] grad = imageGradient(currFrame, new Point((int)warpX, (int)warpY));
                                double dI = templateFrame.Data[y, x, 0] - (contrastScale * affineWarpedFrame.Data[y, x, 0] + brightnessScale);
                                frameMismatchVector[0] += (grad[0] * dI);
                                frameMismatchVector[1] += (grad[1] * dI);
                                frameMismatchVector[2] += (x * grad[0] * dI);
                                frameMismatchVector[3] += (y * grad[0] * dI);
                                frameMismatchVector[4] += (x * grad[1] * dI);
                                frameMismatchVector[5] += (y * grad[1] * dI);
                            }
                        }
                    }  

                // Compute the  LK parameter vector
                LKParameterVector = MatrixOps.matrixVecMult(sumSpatialGradientMatrix, frameMismatchVector);

                // Update the affine warp matrix and displacement
                displacementVector[0] += LKParameterVector[0]; displacementVector[1] += LKParameterVector[1];
                affineWarpMatrix[0, 0] += LKParameterVector[2]; affineWarpMatrix[0, 1] += LKParameterVector[3];
                affineWarpMatrix[1, 0] += LKParameterVector[4]; affineWarpMatrix[1, 1] += LKParameterVector[5];

                iterationCount++;

            } while (iterationCount < maxIterationCount);

        }

        public LucasAndKanadeAffineTracker(Capture source, int nLevels, PointF[] initialVertices, PointF[] vertices, int maxIts)
            : base(source, nLevels, initialVertices, vertices)
        {

            maxIterationCount = maxIts;

            numberOfPyramidLevels = nLevels;

            keyPointFeatureWindowSize = 5;

            searchLength = 5;
        }

     
        public override void trackRegion()
        { 
            alglib.matinvreport rep;
            int info;
            PointF[] warpedBoundingBoxVertices = new PointF[4];
            double[,] spatialGradientMatrix = new double[6, 6];
            double[,] tempMatrix = new double[2, 2];
            double[] frameMismatchVector = new double[6];
            double[] LKParameterVector = new double[6];
            double[,] affineWarpMatrix = new double[2, 2];          // Frame to frame warp matrix
            double[] displacementVector = new double[2];            // Frame to frame displacement
            double[,] startAffineWarpMatrix = new double[2, 2];     // Frame to frame warp matrix
            double[] startDisplacementVector = new double[2];       // Frame to frame displacement
            double[,] totalAffineWarpMatrix = new double[2, 2];     // Frame 0 to current frame warp matrix
            double[] totalDisplacementVector = new double[2];       // Frame 0 to current frame displacement
            double maxDisp;

            sourceVideoFrame = sourceVideo.QueryFrame();

            // Create image to hold warped frame
            warpedFrame = new Image<Gray, Byte>(sourceVideoFrame.Width, sourceVideoFrame.Height);

            // Initialize the total affine parameters
            totalAffineWarpMatrix[0, 0] = totalAffineWarpMatrix[1, 1] = 1.0;
            totalAffineWarpMatrix[1, 0] = totalAffineWarpMatrix[0, 1] = 0.0;
            totalDisplacementVector[0] = totalDisplacementVector[1] = 0.0;

            while (sourceVideoFrame != null)
            {
            
                // Convert to grayscale
                currentFrame = sourceVideoFrame.Convert<Gray, Byte>();
                
                // Compute pyramid
                currentFramePyramid = new ImagePyramid(currentFrame, numberOfPyramidLevels);

                if (initialFrame == null)
                {
                    initialFrame = currentFrame.Clone();

                    initialFramePyramid = new ImagePyramid(initialFrame, numberOfPyramidLevels);
                }

                if (previousFrame != null)
                {
                
                    displayFrame = sourceVideoFrame.Clone().ToBitmap();

                    // Iterate down the pyramid
                    for (int pyramidLevel =  1; pyramidLevel == 1; pyramidLevel--)
                    {
                        // Initialize the affine parameters
                        startAffineWarpMatrix[0, 0] = startAffineWarpMatrix[1, 1] = 1.0;
                        startAffineWarpMatrix[1, 0] = startAffineWarpMatrix[0, 1] = 0.0;
                        startDisplacementVector[0] = startDisplacementVector[1] = 0.0;

                        // Perform the LK alignment
                        align(boundingBoxVertices, previousFramePyramid.PyramidLevel[pyramidLevel], currentFramePyramid.PyramidLevel[pyramidLevel], startAffineWarpMatrix, startDisplacementVector, affineWarpMatrix, displacementVector);


                        // Update the bounding box vertices according to the affine warp matrix and displacement
                        for (int i = 0; i < 4; i++)
                            warpedBoundingBoxVertices[i] = initialBoundingBoxVertices[i];

                        // Update total affine warp parameters
                        updateWarp(totalAffineWarpMatrix, totalDisplacementVector, affineWarpMatrix, displacementVector);

                        // Warp bounding box
                        warpBoundingBox(totalAffineWarpMatrix, totalDisplacementVector, warpedBoundingBoxVertices);
                        
    
                        // Starting from initial bounding box and current total affine parameter estimates, align initial frame
                        MatrixOps.copyMatrix(totalAffineWarpMatrix, startAffineWarpMatrix);
                        startDisplacementVector[0] = totalDisplacementVector[0];
                        startDisplacementVector[1] = totalDisplacementVector[1];

                        align(initialBoundingBoxVertices, initialFramePyramid.PyramidLevel[pyramidLevel], currentFramePyramid.PyramidLevel[pyramidLevel], startAffineWarpMatrix, startDisplacementVector, affineWarpMatrix, displacementVector);

                        for (int i = 0; i < 4; i++)
                        {
                        //    boundingBoxVerticesPrev[i] = boundingBoxVertices[i];
                            boundingBoxVertices[i] = initialBoundingBoxVertices[i];                      
                        }

                        warpBoundingBox(affineWarpMatrix, displacementVector, boundingBoxVertices);

                        maxDisp = boundingBoxDisplacement(boundingBoxVertices, warpedBoundingBoxVertices);

                        triggerDebugMessageEvent("Maxdisp= " + maxDisp);

                         // Test for consistent bounding box  positions
                        if (maxDisp < boundingBoxDisplacementThreshold)
                        {
                            // Update total affine warp parameters
                            MatrixOps.copyMatrix(affineWarpMatrix, totalAffineWarpMatrix);
                            totalDisplacementVector[0] = displacementVector[0];
                            totalDisplacementVector[1] = displacementVector[1];

                            trackingOK = true;
                        }
                        else
                        {
                            trackingOK = false;

                            // Restore bounding box
                            for (int i = 0; i < 4; i++)
                                boundingBoxVertices[i] = warpedBoundingBoxVertices[i];                             
                        } 
                    }       // next pyramid level
                    

                    // Signal for next frame
                    triggerNextFrameEvent();

                }
                       

                // Copy current frame
                previousFrame = currentFrame.Copy();

                previousFramePyramid = new ImagePyramid(previousFrame, numberOfPyramidLevels);

                // Get next frame
                sourceVideoFrame = sourceVideo.QueryFrame();
            }          
        } 
      
    }

    class LucasAndKanadeProjectiveTracker : Tracker
    {
        private ImagePyramid currentFramePyramid, previousFramePyramid, initialFramePyramid;

        private Image<Gray, byte> warpedFrame;

        private int maxIterationCount;

        public LucasAndKanadeProjectiveTracker(Capture source, int nLevels, PointF[] initialVertices, PointF[] vertices, int maxIts)
            : base(source, nLevels, initialVertices, vertices)
        {

            maxIterationCount = maxIts;

            numberOfPyramidLevels = nLevels;
        }

        private void align(PointF[] templateBoundingBoxVertices, Image<Gray, Byte> templateFrame, Image<Gray, Byte> currFrame,
                          double[,] startProjectiveParameterMatrix, double[,] projectiveParameterMatrix)
        {
            alglib.matinvreport rep;
            int info;
            Rectangle templateBoundingRectangle;
            double[,] spatialGradientMatrix = null;
            double[,] sumSpatialGradientMatrix = new double[8, 8];
            double[] frameMismatchVector = new double[8];
            double[] tempVector = new double[8];
            double[] LKParameterVector = new double[8];
            double warpedFrameMean, warpedFrameSD;
            double templateFrameMean, templateFrameSD;
            double contrastScale, brightnessScale;
            double[,] structureMatrix = new double[2, 2];
            double[,] jacobianMatrix = new double[2, 8];
            double[,] tempMatrix1, tempMatrix2;

            // Create image to hold warped frame
            Image<Gray, Byte> projectiveWarpedFrame = new Image<Gray, Byte>(currFrame.Width, currFrame.Height);

            // Determine bounding box rectangle
            templateBoundingRectangle = setBoundingRectangle(templateBoundingBoxVertices);    

            // Copy the initial warp parameters
            for (int i = 0; i < 3; i++)
                for (int j = 0; j < 3; j++)
                    projectiveParameterMatrix[i, j] = startProjectiveParameterMatrix[i, j];

            // Start the LK iterative improvement
            int iterationCount = 0;
            do
            {
                // Warp frame according to current parameter estimates
                for (int x = templateBoundingRectangle.X; x < templateBoundingRectangle.X + templateBoundingRectangle.Width; x++)
                    for (int y = templateBoundingRectangle.Y; y < templateBoundingRectangle.Y + templateBoundingRectangle.Height; y++)
                    {
                        if (insidePolygon(x, y, templateBoundingBoxVertices))
                        {
                            if ((x >= 0) && (x < projectiveWarpedFrame.Width) && (y >= 0) && (y < projectiveWarpedFrame.Height))
                            {
                                double warpX = warpPoint(new PointF(x, y), projectiveParameterMatrix)[0];
                                double warpY = warpPoint(new PointF(x, y), projectiveParameterMatrix)[1];

                                projectiveWarpedFrame[y, x] = new Gray((int)biLinearInterpolate(currFrame, warpX, warpY));
                            }
                        }
                    }

                // Access pixels in current bounding polygon
                // Iterate over each point in the outer bounding box
                // Compute the spatial gradient matrix  

                for (int i = 0; i < 8; i++)
                    for (int j = 0; j < 8; j++)
                        sumSpatialGradientMatrix[i, j] = 0.0;

                // Normalise the mean luminance and contrast of the warped frame to match  the  template frame
                computeStatistics(templateBoundingBoxVertices, templateFrame, out templateFrameMean, out templateFrameSD);

                computeStatistics(templateBoundingBoxVertices, projectiveWarpedFrame, out warpedFrameMean, out warpedFrameSD);

                contrastScale = templateFrameSD / warpedFrameSD;

                brightnessScale = templateFrameMean - contrastScale * warpedFrameMean;

                contrastScale = 1; brightnessScale = 0;

                // Compute the jacobian and the mismatch vector
                for (int i = 0; i < 8; i++)
                    frameMismatchVector[i] = 0.0;

                for (int x = templateBoundingRectangle.X; x < templateBoundingRectangle.X + templateBoundingRectangle.Width; x++)
                    for (int y = templateBoundingRectangle.Y; y < templateBoundingRectangle.Y + templateBoundingRectangle.Height; y++)
                    {
                        if (insidePolygon(x, y, templateBoundingBoxVertices))
                        {
                            if ((x > 0) && (x < currFrame.Width) && (y > 0) && (y < currFrame.Height))
                            {
                                double warpX = warpPoint(new PointF(x, y), projectiveParameterMatrix)[0];
                                double warpY = warpPoint(new PointF(x, y), projectiveParameterMatrix)[1];
                                double[] grad = imageGradient(currFrame, new Point((int)warpX, (int)warpY));
                                structureMatrix[0, 0] = grad[0] * grad[0];
                                structureMatrix[0, 1] = grad[0] * grad[1];
                                structureMatrix[1, 0] = grad[1] * grad[0];
                                structureMatrix[1, 1] = grad[1] * grad[1];

                                double den = projectiveParameterMatrix[2, 0] * x + projectiveParameterMatrix[2, 1] * y + 1;
                                double x1 = (projectiveParameterMatrix[0, 0] * x + projectiveParameterMatrix[0, 1] * y + projectiveParameterMatrix[0, 2]) / den;
                                double y1 = (projectiveParameterMatrix[1, 0] * x + projectiveParameterMatrix[1, 1] * y + projectiveParameterMatrix[1, 2]) / den;

                                jacobianMatrix[0, 0] = x/den; jacobianMatrix[1, 0] = 0;
                                jacobianMatrix[0, 1] = y/den; jacobianMatrix[1, 1] = 0;
                                jacobianMatrix[0, 2] = 1/den; jacobianMatrix[1, 2] = 0;
                                jacobianMatrix[0, 3] = 0; jacobianMatrix[1, 3] = x/den;
                                jacobianMatrix[0, 4] = 0; jacobianMatrix[1, 4] = y/den;
                                jacobianMatrix[0, 5] = 0; jacobianMatrix[1, 5] = 1/den;
                                jacobianMatrix[0, 6] = (-x * x1)/den; jacobianMatrix[1, 6] = (-x * y1)/den;
                                jacobianMatrix[0, 7] = (-y * x1)/den; jacobianMatrix[1, 7] = (-y * y1)/den;

                                tempMatrix1 = MatrixOps.matrixMult(structureMatrix, jacobianMatrix);

                                tempMatrix2 = MatrixOps.matrixTranspose(jacobianMatrix);

                                spatialGradientMatrix = MatrixOps.matrixMult(tempMatrix2, tempMatrix1);

                                MatrixOps.matrixSum(sumSpatialGradientMatrix, spatialGradientMatrix);

                                double dI = templateFrame.Data[y, x, 0] - (contrastScale * projectiveWarpedFrame.Data[y, x, 0] + brightnessScale);

                                frameMismatchVector[0] += (x * grad[0] * dI) / den;
                                frameMismatchVector[1] += (y * grad[0] * dI) / den;
                                frameMismatchVector[2] += (grad[0] * dI) / den;
                                frameMismatchVector[3] += (x * grad[1] * dI) / den;
                                frameMismatchVector[4] += (y * grad[1] * dI) / den;
                                frameMismatchVector[5] += (grad[1] * dI) / den;
                                frameMismatchVector[6] += ((-x1 * x * grad[0] - y1 * x * grad[1]) * dI) / den;
                                frameMismatchVector[7] += ((-x1 * y * grad[0] - y1 * y * grad[1]) * dI) / den;
                            }
                        }
                    }

                // Invert the spatial gradient matrix
                alglib.rmatrixinverse(ref sumSpatialGradientMatrix, out info, out rep);           

               /* for (int x = templateBoundingRectangle.X; x < templateBoundingRectangle.X + templateBoundingRectangle.Width; x++)
                    for (int y = templateBoundingRectangle.Y; y < templateBoundingRectangle.Y + templateBoundingRectangle.Height; y++)
                    {
                        if (insidePolygon(x, y, templateBoundingBoxVertices))
                        {
                            if ((x > 0) && (x < currFrame.Width) && (y > 0) && (y < currFrame.Height))
                            {
                                double warpX = warpPoint(new PointF(x, y), projectiveParameterMatrix)[0];
                                double warpY = warpPoint(new PointF(x, y), projectiveParameterMatrix)[1];
                                double[] grad = imageGradient(currFrame, new Point((int)warpX, (int)warpY));
                                double dI = templateFrame.Data[y, x, 0] - (contrastScale * projectiveWarpedFrame.Data[y, x, 0] + brightnessScale);
                                double den = projectiveParameterMatrix[2, 0] * x + projectiveParameterMatrix[2, 1] * y + 1;
                                double x1 = (projectiveParameterMatrix[0, 0] * x + projectiveParameterMatrix[0, 1] * y + projectiveParameterMatrix[0, 2]) / den;
                                double y1 = (projectiveParameterMatrix[1, 0] * x + projectiveParameterMatrix[1, 1] * y + projectiveParameterMatrix[1, 2]) / den;
                                
                                frameMismatchVector[0] += (x * grad[0] * dI)/den;
                                frameMismatchVector[1] += (y * grad[0] * dI)/den;
                                frameMismatchVector[2] += (grad[0] * dI)/den;
                                frameMismatchVector[3] += (x * grad[1] * dI)/den;
                                frameMismatchVector[4] += (y * grad[1] * dI)/den;
                                frameMismatchVector[5] += (grad[1] * dI)/den;
                                frameMismatchVector[6] += ((-x1 * x * grad[0] - y1 * x * grad[1])*dI)/den;
                                frameMismatchVector[7] += ((-x1 * y * grad[0] - y1 * y * grad[1])*dI)/den;
                            }
                        }
                    } */
           
                LKParameterVector = MatrixOps.matrixVecMult(sumSpatialGradientMatrix, frameMismatchVector);

                projectiveParameterMatrix[0, 0] += LKParameterVector[0];
                projectiveParameterMatrix[0, 1] += LKParameterVector[1];
                projectiveParameterMatrix[0, 2] += LKParameterVector[2];
                projectiveParameterMatrix[1, 0] += LKParameterVector[3];
                projectiveParameterMatrix[1, 1] += LKParameterVector[4];
                projectiveParameterMatrix[1, 2] += LKParameterVector[5];
                projectiveParameterMatrix[2, 0] += LKParameterVector[6];
                projectiveParameterMatrix[2, 1] += LKParameterVector[7];
                projectiveParameterMatrix[2, 2] = 1;

                iterationCount++;

            } while (iterationCount < maxIterationCount);

        }

        public override void trackRegion()
        {
            alglib.matinvreport rep;
            int info;
            double[,] totalProjectiveParameterMatrix = new double[3,3];
            double[,] startProjectiveParameterMatrix = new double[3,3];
            double[,] projectiveParameterMatrix = new double[3,3];
            PointF[] warpedBoundingBoxVertices = new PointF[4];
            double maxDisp;

            sourceVideoFrame = sourceVideo.QueryFrame();

            // Create image to hold warped frame
            warpedFrame = new Image<Gray, Byte>(sourceVideoFrame.Width, sourceVideoFrame.Height);

            // Initialize the total projective parameters
            for (int i = 0; i < 3; i++)
                for (int j = 0; j < 3; j++)
                    totalProjectiveParameterMatrix[i, j] = 0;

            totalProjectiveParameterMatrix[0, 0] = totalProjectiveParameterMatrix[1, 1] = totalProjectiveParameterMatrix[2, 2] = 1.0;

            int frameNumber = 0;
            while (sourceVideoFrame != null)
            {
  
                frameNumber++;

                // Convert to grayscale
                currentFrame = sourceVideoFrame.Convert<Gray, Byte>();

                // Compute pyramid
                currentFramePyramid = new ImagePyramid(currentFrame, numberOfPyramidLevels);

                if (initialFrame == null)
                {
                    initialFrame = currentFrame.Clone();

                    initialFramePyramid = new ImagePyramid(initialFrame, numberOfPyramidLevels);
                }

                if (previousFrame != null)
                {

                    displayFrame = sourceVideoFrame.Clone().ToBitmap();

                    // Iterate down the pyramid
                    for (int pyramidLevel = 1; pyramidLevel == 1; pyramidLevel--)
                    {
                        // Initialize the start projective parameters
                        for (int i = 0; i < 3; i++)
                            for (int j = 0; j < 3; j++)
                                startProjectiveParameterMatrix[i, j] = 0;

                        startProjectiveParameterMatrix[0, 0] = 1.0;
                        startProjectiveParameterMatrix[1, 1] = 1.0;
                        startProjectiveParameterMatrix[2, 2] = 1.0;
                        
                        // Perform the LK alignment
                        align(boundingBoxVertices, previousFramePyramid.PyramidLevel[pyramidLevel], currentFramePyramid.PyramidLevel[pyramidLevel], startProjectiveParameterMatrix, projectiveParameterMatrix);

                       // Update the bounding box vertices 
                       for (int i = 0; i < 4; i++)
                            warpedBoundingBoxVertices[i] = initialBoundingBoxVertices[i];

                        updateWarp(totalProjectiveParameterMatrix, projectiveParameterMatrix); 

                        // Warp bounding box
                        warpBoundingBox(totalProjectiveParameterMatrix, warpedBoundingBoxVertices);


                        // Starting from initial bounding box and current total affine parameter estimates, align initial frame
                        
                        MatrixOps.copyMatrix(totalProjectiveParameterMatrix, startProjectiveParameterMatrix);
                       
                        // Perform the LK alignment
                        align(initialBoundingBoxVertices, initialFramePyramid.PyramidLevel[pyramidLevel], currentFramePyramid.PyramidLevel[pyramidLevel], startProjectiveParameterMatrix,  projectiveParameterMatrix);

                        for (int i = 0; i < 4; i++)
                            boundingBoxVertices[i] = initialBoundingBoxVertices[i];  

                        warpBoundingBox(projectiveParameterMatrix, boundingBoxVertices);

                        maxDisp = boundingBoxDisplacement(boundingBoxVertices, warpedBoundingBoxVertices);

                        triggerDebugMessageEvent("Maxdisp= " + maxDisp);

                        // Test for consistent bounding box  positions
                        if (maxDisp < boundingBoxDisplacementThreshold)
                        {
                            // Update total affine warp parameters
                            MatrixOps.copyMatrix(projectiveParameterMatrix, totalProjectiveParameterMatrix);
                            
                            trackingOK = true;
                        }
                        else
                        {
                            trackingOK = false;

                            // Restore bounding box
                            for (int i = 0; i < 4; i++)
                                boundingBoxVertices[i] = warpedBoundingBoxVertices[i];
                        }    
                    }       // next pyramid level

                    // Signal for next frame
                    triggerNextFrameEvent();
                }

                // Copy current frame
                previousFrame = currentFrame.Copy();

                previousFramePyramid = new ImagePyramid(previousFrame, numberOfPyramidLevels);

                // Get next frame
                sourceVideoFrame = sourceVideo.QueryFrame();

            }
        }
    }
}
